Attribute VB_Name = "modMain"
' /*******************************************************************************
'   modMain.bas within vbSpec.vbp
'
'   Public variable declarations, startup and initialization code,
'   and routines for loading the ROM and snapshots
'
'   Author: Chris Cowley <ccowley@grok.co.uk>
'
'   Copyright (C)1999-2000 Grok Developments Ltd.
'   http://www.grok.co.uk/
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/

Option Explicit

Public glTstatesPerInterrupt As Long
Public Parity(256) As Boolean
Public glInterruptTimer As Long

' // Main Z80 registers //
Public regA As Long
Public regHL As Long
Public regB As Long
Public regC As Long
Public regDE As Long

' // Z80 Flags
Public fS As Boolean
Public fZ As Boolean
Public f5 As Boolean
Public fH As Boolean
Public f3 As Boolean
Public fPV As Boolean
Public fN As Boolean
Public fC As Boolean

' // Flag bit positions
Public Const F_C = 1
Public Const F_N = 2
Public Const F_PV = 4
Public Const F_3 = 8
Public Const F_H = 16
Public Const F_5 = 32
Public Const F_Z = 64
Public Const F_S = 128

' // Alternate registers //
Public regAF_ As Long
Public regHL_ As Long
Public regBC_ As Long
Public regDE_ As Long

' // Index registers  - ID used as temp for ix/iy
Public regIX As Long
Public regIY As Long
Public regID As Long

' // Stack pointer and program counter
Public regSP As Long
Public regPC As Long

' // Interrupt registers and flip-flops, and refresh registers
Public intI As Long
Public intR As Long
Public intRTemp As Long
Public intIFF1 As Boolean
Public intIFF2 As Boolean
Public intIM As Long

' // Addressable Memory (16KB ROM at 0x0000, 48KB RAM at 0x4000)
' // We also place a shadow copy of the ROM above the addressable memory space
' // at 0x10000 so we don't have to check for regPC > 65535 at each instruction
Public mem(82370) As Byte

' // Screen things
Public Const nPixelsWide = 256
Public Const nPixelsHigh = 192
Public Const nCharsWide = 32
Public Const nCharsHigh = 24
Public Const FirstAttr = nPixelsHigh * nCharsWide
Public Const lastAddr = FirstAttr + (nCharsHigh * nCharsWide)

Public ScrnLines(nPixelsHigh - 1, 32) As Boolean
Public ScrnNeedRepaint As Boolean
Public bFlashInverse As Boolean
Public glScreenMem(191, 31) As Long

' // Array of colour values (speeds up screen painting by avoiding
' // multiple calls to RGB() )
Public glBrightColor(0 To 7) As Long
Public glNormalColor(0 To 7) As Long

' // Global picDisplay variable to speed things up
Public gpicDisplay As PictureBox

' // Interrupts/Screen refreshing
Public interruptCounter As Long
Public refreshRate As Long
Public glInterruptDelay As Long

' // Keypresses
Public keyB_SPC As Long
Public keyH_ENT As Long
Public keyY_P As Long
Public key6_0 As Long
Public key1_5 As Long
Public keyQ_T As Long
Public keyA_G As Long
Public keyCAPS_V As Long

' // Sadly, I needed to use these high res timer functions to precisely control the
' // speed of emulation. I had hoped to do this without resorting to any API calls :(
Public Declare Function timeGetTime Lib "winmm.dll" () As Long
Public Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Long) As Long
Public Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Long) As Long
Public Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

' // If build with the USE_WINAPI compiler directive defined, vbSpec uses Windows API
' // functions to paint the display. This is faster than raw VB code, and provides the
' // option for selecting scaled display sizes (double and triple size, and so on).
#If USE_WINAPI Then
    Public Const SRCCOPY = &HCC0020 ' (DWORD) dest = source
    Public Declare Function StretchDIBits Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal dx As Long, ByVal dy As Long, ByVal SrcX As Long, ByVal SrcY As Long, ByVal wSrcWidth As Long, ByVal wSrcHeight As Long, lpBits As Any, lpBitsInfo As BITMAPINFO, ByVal wUsage As Long, ByVal dwRop As Long) As Long
    Public Declare Function GetSystemMetrics Lib "user32" (ByVal nIndex As Long) As Long
    Public Const SM_CYCAPTION = 4
    Public Const SM_CYMENU = 15
    Public Const SM_CXFRAME = 32
    Public Const SM_CYFRAME = 33
    
    Public Type BITMAPINFOHEADER '40 bytes
            biSize As Long
            biWidth As Long
            biHeight As Long
            biPlanes As Integer
            biBitCount As Integer
            biCompression As Long
            biSizeImage As Long
            biXPelsPerMeter As Long
            biYPelsPerMeter As Long
            biClrUsed As Long
            biClrImportant As Long
    End Type
    Public Type RGBQUAD
            rgbBlue As Byte
            rgbGreen As Byte
            rgbRed As Byte
            rgbReserved As Byte
    End Type
    Public Type BITMAPINFO
            bmiHeader As BITMAPINFOHEADER
            bmiColors(16) As RGBQUAD
    End Type
    Public Const BI_RGB = 0&
    Public Const DIB_RGB_COLORS = 0 '  color table in RGBs
    
    Public bmiBuffer As BITMAPINFO
    Public cBufferBits(49152) As Byte
    
    Public glDisplayHeight As Long
    Public glDisplayWidth As Long
    Public glDisplayVSource As Long ' // Set to glDisplayHeight - 1 to improve display speed
    Public glDisplayVSize As Long   ' // Set to -glDisplayHeight to improve display speed
#End If

' // ShellExecute is used for the clickable web URL in the "About..." dialog, not
' //the actual emulator itself
Public Declare Function ShellExecute Lib "shell32.dll" Alias "ShellExecuteA" (ByVal hwnd As Long, ByVal lpOperation As String, ByVal lpFile As String, ByVal lpParameters As String, ByVal lpDirectory As String, ByVal nShowCmd As Long) As Long



Public Sub CreateScreenBuffer()
    #If USE_WINAPI Then
        bmiBuffer.bmiHeader.biSize = Len(bmiBuffer.bmiHeader)
        bmiBuffer.bmiHeader.biWidth = 256
        bmiBuffer.bmiHeader.biHeight = 192
        bmiBuffer.bmiHeader.biPlanes = 1
        bmiBuffer.bmiHeader.biBitCount = 8
        bmiBuffer.bmiHeader.biCompression = BI_RGB
        bmiBuffer.bmiHeader.biSizeImage = 0
        bmiBuffer.bmiHeader.biXPelsPerMeter = 200
        bmiBuffer.bmiHeader.biYPelsPerMeter = 200
        bmiBuffer.bmiHeader.biClrUsed = 16
        bmiBuffer.bmiHeader.biClrImportant = 16
    #End If
End Sub

Function GetFilePart(ByVal sFilename As String) As String
    If InStr(sFilename, ":") > 0 Then sFilename = Mid$(sFilename, InStr(sFilename, ":") + 1)
    Do While InStr(sFilename, "\") > 0
        sFilename = Mid$(sFilename, InStr(sFilename, "\") + 1)
    Loop
    GetFilePart = sFilename
End Function

Sub initParity()
    Dim lCounter As Long, j As Byte, p As Boolean
    
    For lCounter = 0 To 255
        p = True
        For j = 0 To 7
            If (lCounter And (2 ^ j)) <> 0 Then p = Not p
        Next j
        Parity(lCounter) = p
    Next lCounter
End Sub

Sub initscreen()
    Dim i As Long, X As Long
    
    ScrnNeedRepaint = True
    For i = 0 To 191
        ScrnLines(i, 32) = True
        For X = 0 To 31
            ScrnLines(i, X) = True
        Next X
    Next i
End Sub

Public Sub InitScreenMemTable()
    Dim X As Long, Y As Long
    
    For Y = 0 To 191
        For X = 0 To 31
            glScreenMem(Y, X) = ((((Y \ 8) * 32) + (Y Mod 8) * 256) + ((Y \ 64) * 2048) - (Y \ 64) * 256) + 16384 + X
        Next X
    Next Y
End Sub


Sub LoadROM(Optional sROMFile As String = "spectrum.rom")
    Dim hFile As Long, sROM As String, iCounter As Long
    
    On Error Resume Next
    
    If Dir$(sROMFile) = "" Then
        MsgBox "The ROM image file '" & sROMFile & "' could not be found.", vbExclamation Or vbOKOnly
        Exit Sub
    End If
    
    Err.Clear
    hFile = FreeFile
    Open sROMFile For Binary As hFile
    If Err.Number <> 0 Then
        MsgBox "Unable to open ROM image file: " & sROMFile, vbExclamation Or vbOKOnly
        Close hFile
        Exit Sub
    End If
    
    ' // Read the ROM image into sROM
    Err.Clear
    sROM = Input(16384, #hFile)
    Close hFile
        
    If Err.Number <> 0 Then
        MsgBox "An error ocurred whilst reading the ROM image file: " & sROMFile, vbExclamation Or vbOKOnly
        Exit Sub
    End If
    
    ' // Copy the ROM in to the low 16KB of addressable memory, and also place a
    ' // shadow copy at 65536 - this means we don't have to check for regPC > 65535
    ' // which improves performance
    For iCounter = 1 To 16384
        mem(iCounter - 1) = Asc(Mid$(sROM, iCounter, 1))
        mem(iCounter + 65535) = mem(iCounter - 1)
    Next iCounter
End Sub

Public Sub LoadSNASnap(sFilename As String)
    Dim hFile As Long, sData As String, iCounter As Long
    
    hFile = FreeFile
    Open sFilename For Binary As hFile
    
    sData = Input(1, #hFile)
    intI = Asc(sData)
    sData = Input(2, #hFile)
    regHL_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regDE_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regBC_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regAF_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(2, #hFile)
    regHL = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regDE = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    setBC Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regIY = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regIX = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(1, #hFile)
    If Asc(sData) And 4 Then
        intIFF1 = True
        intIFF2 = True
    Else
        intIFF1 = False
        intIFF2 = False
    End If
    
    sData = Input(1, #hFile)
    intR = Asc(sData)
    intRTemp = intR
    
    sData = Input(2, #hFile)
    setAF Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(2, #hFile)
    regSP = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))

    sData = Input(1, #hFile)
    intIM = Asc(sData)
    
    sData = Input(1, #hFile)
    ' // Border color
    glNewBorder = (Asc(sData) And &H7&)
    
    Err.Clear
    sData = Input(49153, #hFile)
    Close hFile
       
    For iCounter = 1 To 49152
        mem(iCounter + 16383) = Asc(Mid$(sData, iCounter, 1))
    Next iCounter
    
    initscreen
    
    ' // Set the initial border color
    Select Case glNewBorder
    Case 0
        frmMainWnd.BackColor = 0&
    Case 1
        frmMainWnd.BackColor = RGB(0, 0, 192)
    Case 2
        frmMainWnd.BackColor = RGB(192, 0, 0)
    Case 3
        frmMainWnd.BackColor = RGB(192, 0, 192)
    Case 4
        frmMainWnd.BackColor = RGB(0, 192, 0)
    Case 5
        frmMainWnd.BackColor = RGB(0, 192, 192)
    Case 6
        frmMainWnd.BackColor = RGB(192, 192, 0)
    Case 7
        frmMainWnd.BackColor = RGB(192, 192, 192)
    End Select
    
    frmMainWnd.Caption = App.ProductName & " - " & GetFilePart(sFilename)
    
    screenPaint
    resetKeyboard
    poppc
End Sub


Public Sub LoadZ80Snap(sFilename As String)
    Dim hFile As Long, sData As String, iCounter As Long
    Dim bCompressed As Boolean
    
    hFile = FreeFile
    Open sFilename For Binary As hFile
    
    ' byte 0 - A register
    sData = Input(1, #hFile)
    regA = Asc(sData)
    ' byte 1 - F register
    sData = Input(1, #hFile)
    setF Asc(sData)
    ' bytes 2 + 3 - BC register pair (C first, then B)
    sData = Input(2, #hFile)
    setBC Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 4 + 5 - HL register pair
    sData = Input(2, #hFile)
    regHL = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 6 + 7 - PC (this is zero for v2.x or v3.0 Z80 files)
    sData = Input(2, #hFile)
    regPC = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 8 + 9 - SP
    sData = Input(2, #hFile)
    regSP = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 10 - Interrupt register
    sData = Input(1, #hFile)
    intI = Asc(sData)
    ' byte 11 - Refresh register
    sData = Input(1, #hFile)
    intR = Asc(sData)
    intRTemp = intR
    
    ' byte 12 - bitfield
    sData = Input(1, #hFile)
    ' if byte 12 = 255 then it must be treated as if it = 1, for compatibility with other emulators
    If Asc(sData) = 255 Then sData = Chr$(1)
    ' bit 0 - bit 7 of R
    If (Asc(sData) And 1) = 0 Then intR = intR And &H80& Else intR = intR And Not (128)
    ' bits 1,2 and 3 - border color
    glNewBorder = (Asc(sData) And 14) \ 2
    ' bit 4 - 1 if SamROM switched in (we don't care about this!)
    ' bit 5 - if 1 and PC<>0 then the snapshot is compressed using the
    '         rudimentary Z80 run-length encoding scheme
    If (Asc(sData) And &H20&) Then bCompressed = True
    ' bits 6 + 7 - no meaning

    ' bytes 13 + 14 - DE register pair
    sData = Input(2, #hFile)
    regDE = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 15 + 16 - BC' register pair
    sData = Input(2, #hFile)
    regBC_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 17 + 18 - DE' register pair
    sData = Input(2, #hFile)
    regDE_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 19 + 20 - HL' register pair
    sData = Input(2, #hFile)
    regHL_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 21 + 22 - AF' register pair (A first then F - not Z80 byte order!!)
    sData = Input(2, #hFile)
    regAF_ = Asc(Left$(sData, 1)) * 256& + Asc(Right$(sData, 1))
    ' byte 23 + 24 - IY register pair
    sData = Input(2, #hFile)
    regIY = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 25 + 26 - IY register pair
    sData = Input(2, #hFile)
    regIX = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 27 - Interrupt flipflop (0=DI, else EI)
    sData = Input(1, #hFile)
    If Asc(sData) = 0 Then
        intIFF1 = False
        intIFF2 = False
    Else
        intIFF1 = True
        intIFF2 = True
    End If
    ' byte 28 - IFF2 (ignored)
    sData = Input(1, #hFile)
    ' byte 29 - Interrupt mode (bits 2 - 7 contain info about joystick modes etc, which we ignore)
    sData = Input(1, #hFile)
    intIM = Asc(sData) And 3
    
    If regPC = 0 Then
        ' This is a V2 or V3 Z80 file
        ReadZ80V2orV3Snap hFile
        Close hFile
    Else
        ' // PC<>0, so lets check to see if this is a compressed V1 Z80 file
        If bCompressed Then
            ' Uncompress the RAM data
            ReadZ80V1Snap hFile
            Close hFile
        Else
            ' // Uncompressed Z80 file
            sData = Input(49153, #hFile)
            Close hFile
            
            ' // Copy the RAM data to addressable memory space
            For iCounter = 1 To 49152
                mem(iCounter + 16383) = Asc(Mid$(sData, iCounter, 1))
            Next iCounter
        End If
    End If
    
    initscreen
    
    ' // Set the initial border color
    Select Case glNewBorder
    Case 0
        frmMainWnd.BackColor = 0&
    Case 1
        frmMainWnd.BackColor = RGB(0, 0, 192)
    Case 2
        frmMainWnd.BackColor = RGB(192, 0, 0)
    Case 3
        frmMainWnd.BackColor = RGB(192, 0, 192)
    Case 4
        frmMainWnd.BackColor = RGB(0, 192, 0)
    Case 5
        frmMainWnd.BackColor = RGB(0, 192, 192)
    Case 6
        frmMainWnd.BackColor = RGB(192, 192, 0)
    Case 7
        frmMainWnd.BackColor = RGB(192, 192, 192)
    End Select

    frmMainWnd.Caption = App.ProductName & " - " & GetFilePart(sFilename)
    
    screenPaint
    resetKeyboard
    
    gpicDisplay.REFRESH
    DoEvents
End Sub


Sub Main()
    InitColorArrays
    InitScreenMemTable
      
    frmMainWnd.Show
    
    Set gpicDisplay = frmMainWnd.picDisplay
    
    #If USE_WINAPI Then
        CreateScreenBuffer
        glDisplayWidth = val(GetSetting("Grok", "vbSpec", "DisplayWidth", "256"))
        glDisplayHeight = val(GetSetting("Grok", "vbSpec", "DisplayHeight", "192"))
        SetDisplaySize glDisplayWidth, glDisplayHeight
    #Else
        ' // Hide the "Display..." Options item, as it can't be changed unless we're
        ' // using the fast WINAPI method of screen painting
        frmMainWnd.mnuOptions(2).Visible = False
    #End If
    
    refreshRate = 1
    glInterruptDelay = val(GetSetting("Grok", "vbSpec", "InterruptDelay", "20"))
    
    ' // Initialize everything
    intIFF1 = True
    intIFF2 = True
    intIM = 2
    initParity
    
    ' // 3.5 MHz
    glTstatesPerInterrupt = 69888
    ' there are 69888 tstates per interrupt on a real spectrum
    
    ' // load the ROM image into memory
    LoadROM
    initscreen
    resetKeyboard
    
    timeBeginPeriod 1
    
    ' // Begin executing the ROM code at 0x0000
    glInterruptTimer = timeGetTime() + 20
    execute
End Sub

Public Sub InitColorArrays()
#If USE_WINAPI Then
    bmiBuffer.bmiColors(0).rgbRed = 0
    bmiBuffer.bmiColors(0).rgbGreen = 0
    bmiBuffer.bmiColors(0).rgbBlue = 0
    
    bmiBuffer.bmiColors(1).rgbRed = 0
    bmiBuffer.bmiColors(1).rgbGreen = 0
    bmiBuffer.bmiColors(1).rgbBlue = 192
    
    bmiBuffer.bmiColors(2).rgbRed = 192
    bmiBuffer.bmiColors(2).rgbGreen = 0
    bmiBuffer.bmiColors(2).rgbBlue = 0
    
    bmiBuffer.bmiColors(3).rgbRed = 192
    bmiBuffer.bmiColors(3).rgbGreen = 0
    bmiBuffer.bmiColors(3).rgbBlue = 192
    
    bmiBuffer.bmiColors(4).rgbRed = 0
    bmiBuffer.bmiColors(4).rgbGreen = 192
    bmiBuffer.bmiColors(4).rgbBlue = 0
    
    bmiBuffer.bmiColors(5).rgbRed = 0
    bmiBuffer.bmiColors(5).rgbGreen = 192
    bmiBuffer.bmiColors(5).rgbBlue = 192
    
    bmiBuffer.bmiColors(6).rgbRed = 192
    bmiBuffer.bmiColors(6).rgbGreen = 192
    bmiBuffer.bmiColors(6).rgbBlue = 0
    
    bmiBuffer.bmiColors(7).rgbRed = 192
    bmiBuffer.bmiColors(7).rgbGreen = 192
    bmiBuffer.bmiColors(7).rgbBlue = 192
    
    bmiBuffer.bmiColors(8).rgbRed = 0
    bmiBuffer.bmiColors(8).rgbGreen = 0
    bmiBuffer.bmiColors(8).rgbBlue = 0
    
    bmiBuffer.bmiColors(9).rgbRed = 0
    bmiBuffer.bmiColors(9).rgbGreen = 0
    bmiBuffer.bmiColors(9).rgbBlue = 255
    
    bmiBuffer.bmiColors(10).rgbRed = 255
    bmiBuffer.bmiColors(10).rgbGreen = 0
    bmiBuffer.bmiColors(10).rgbBlue = 0
    
    bmiBuffer.bmiColors(11).rgbRed = 255
    bmiBuffer.bmiColors(11).rgbGreen = 0
    bmiBuffer.bmiColors(11).rgbBlue = 255
    
    bmiBuffer.bmiColors(12).rgbRed = 0
    bmiBuffer.bmiColors(12).rgbGreen = 255
    bmiBuffer.bmiColors(12).rgbBlue = 0
    
    bmiBuffer.bmiColors(13).rgbRed = 0
    bmiBuffer.bmiColors(13).rgbGreen = 255
    bmiBuffer.bmiColors(13).rgbBlue = 255
    
    bmiBuffer.bmiColors(14).rgbRed = 255
    bmiBuffer.bmiColors(14).rgbGreen = 255
    bmiBuffer.bmiColors(14).rgbBlue = 0
    
    bmiBuffer.bmiColors(15).rgbRed = 255
    bmiBuffer.bmiColors(15).rgbGreen = 255
    bmiBuffer.bmiColors(15).rgbBlue = 255
#End If
    glBrightColor(0) = 0
    glBrightColor(1) = RGB(0, 0, 255)
    glBrightColor(2) = RGB(255, 0, 0)
    glBrightColor(3) = RGB(255, 0, 255)
    glBrightColor(4) = RGB(0, 255, 0)
    glBrightColor(5) = RGB(0, 255, 255)
    glBrightColor(6) = RGB(255, 255, 0)
    glBrightColor(7) = RGB(255, 255, 255)
    glNormalColor(0) = 0
    glNormalColor(1) = RGB(0, 0, 192)
    glNormalColor(2) = RGB(192, 0, 0)
    glNormalColor(3) = RGB(192, 0, 192)
    glNormalColor(4) = RGB(0, 192, 0)
    glNormalColor(5) = RGB(0, 192, 192)
    glNormalColor(6) = RGB(192, 192, 0)
    glNormalColor(7) = RGB(192, 192, 192)
End Sub
Private Sub ReadZ80V1Snap(hFile As Long)
    Dim lDataLen As Long, sData As String, lBlockLen As Long
    Dim lCounter As Long, lMemPos As Long, lBlockCounter As Long
 
    lDataLen = LOF(hFile) - Seek(hFile) + 1
    ' // read the compressed data into sData
    sData = Input(lDataLen, #hFile)
        
    ' // Uncompress the block to memory
    lCounter = 1
    lMemPos = 16384
    Do
        If Asc(Mid$(sData, lCounter, 1)) = &HED& Then
            If Asc(Mid$(sData, lCounter + 1, 1)) = &HED& Then
                ' // This is an encoded block
                lCounter = lCounter + 2
                lBlockLen = Asc(Mid$(sData, lCounter, 1))
                lCounter = lCounter + 1
                For lBlockCounter = 1 To lBlockLen
                    mem(lMemPos) = Asc(Mid$(sData, lCounter, 1))
                    lMemPos = lMemPos + 1
                Next lBlockCounter
            Else
                ' // Just a single ED, write it out
                mem(lMemPos) = &HED&
                lMemPos = lMemPos + 1
            End If
        Else
            mem(lMemPos) = Asc(Mid$(sData, lCounter, 1))
            lMemPos = lMemPos + 1
        End If
        lCounter = lCounter + 1
    Loop Until lCounter > Len(sData) - 4
    
    If Mid$(sData, lCounter, 4) <> Chr$(0) & Chr$(&HED) & Chr$(&HED) & Chr$(0) Then
        MsgBox "Error in compressed Z80 file. Block end marker 0x00EDED00 is not present."
    End If

End Sub

Private Sub ReadZ80V2orV3Snap(hFile As Long)
    Dim lHeaderLen As Long, sData As String
    Dim lCounter As Long, bHardwareSupported As Boolean
    Dim lMemPos As Long, lBlockCounter As Long
    
    sData = Input(2, #hFile)
    lHeaderLen = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    ' // offset 32 - 2 bytes - PC
    If lCounter < lHeaderLen Then
        sData = Input(2, #hFile)
        regPC = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
        lCounter = lCounter + 2
    End If
    
    ' // offset 34 - 1 byte - hardware mode
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        Select Case Asc(sData)
        Case 0 ' // 48K spectrum
            bHardwareSupported = True
        Case 1 ' // 48K spectrum + Interface 1
            bHardwareSupported = True
        Case 2 ' // SamROM
            sData = "SamRam"
            bHardwareSupported = False
        Case 3
            If lHeaderLen = 23 Then
                sData = "128K Spectrum"
                bHardwareSupported = False
            Else
                sData = "48K Spectrum + M.G.T."
                bHardwareSupported = True
            End If
        Case 4
            bHardwareSupported = False
            If lHeaderLen = 23 Then
                sData = "128K Spectrum + Interface 1"
            Else
                sData = "128K Spectrum"
            End If
        Case 5
            bHardwareSupported = False
            sData = "128K Spectrum + Interface 1"
        Case 6
            bHardwareSupported = True
            sData = "128K Spectrum + M.G.T."
        Case Else
            bHardwareSupported = False
            sData = "Unknown hardware platform"
        End Select
        
        If bHardwareSupported = False Then
            MsgBox "vbSpec does not support the required hardware platform (" & sData & ") for this snapshot. This snapshot is unlikely to function correctly in vbSpec."
        End If
        
        lCounter = lCounter + 1
    End If
    
    ' // offset 35 - 1 byte - last out to 0x7FFD - not required for 48K spectrum
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
    End If
    
    ' // offset 36 - 1 byte - 0xFF if Interface 1 ROM is paged in
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
        
        If Asc(sData) = 255 Then
            MsgBox "This snapshot was saved with an Interface 1 ROM paged in. It may not run correctly in vbSpec."
        End If
    End If
    
    ' // offset 37 - 1 byte (bit 0: 1=intR emulation on, bit 1: 1=LDIR emulation on
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
    End If
    
    ' // offset 38 - Last out to 0xFFFD (+2/+3 sound chip register number)
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
    End If
    
    ' // offset 39 - 16 bytes - contents of the sound chip registers
    If lCounter < lHeaderLen Then
        sData = Input(16, #hFile)
        lCounter = lCounter + 16
    End If
    
    ' // read the remaining bytes of the header (we don't care what information they hold)
    If lCounter < lHeaderLen Then
        sData = Input(lHeaderLen - lCounter, #hFile)
    End If
    
    Do
        ' // read a block
        sData = Input(2, #hFile)
        If EOF(hFile) Then Exit Do
        lHeaderLen = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
        sData = Input(1, #hFile)
        Select Case Asc(sData)
        Case 0 ' // Spectrum ROM
            lMemPos = 0
        Case 1 ' // Interface 1 ROM, or similar (we discard these blocks)
            lMemPos = -1
        Case 2 ' // 128K ROM (reset) (we discard these blocks)
            lMemPos = -1
        Case 3 ' // Page 0 (not used by 48K Spectrum)
            lMemPos = -1
        Case 4 ' // Page 1 RAM at 0x8000
            lMemPos = &H8000&
        Case 5 ' // Page 2 RAM at 0xC000
            lMemPos = &HC000&
        Case 6 ' // Page 3 (not used by 48K Spectrum)
            lMemPos = -1
        Case 7 ' // Page 4 (not used by 48K Spectrum)
            lMemPos = -1
        Case 8 ' // Page 5 RAM at 0x4000
            lMemPos = &H4000&
        Case 9 ' // Page 6 (not used by 48K Spectrum)
            lMemPos = -1
        Case 10 ' // Page 7 (not used by 48K Spectrum)
            lMemPos = -1
        Case 11 ' // Multiface ROM
        End Select
        sData = Input(lHeaderLen, #hFile)
        
        If lMemPos <> -1 Then
            If lHeaderLen = &HFFFF& Then
                ' Not a compressed block, just copy it straight into RAM
                For lCounter = 1 To 16384
                    mem(lCounter - 1 + lMemPos) = Asc(Mid$(sData, lCounter, 1))
                Next lCounter
            Else
                ' // Uncompress the block to memory
                lCounter = 1
                Do
                    If Asc(Mid$(sData, lCounter, 1)) = &HED& Then
                        If Asc(Mid$(sData, lCounter + 1, 1)) = &HED& Then
                            ' // This is an encoded block
                            lCounter = lCounter + 2
                            lHeaderLen = Asc(Mid$(sData, lCounter, 1))
                            lCounter = lCounter + 1
                            For lBlockCounter = 1 To lHeaderLen
                                mem(lMemPos) = Asc(Mid$(sData, lCounter, 1))
                                lMemPos = lMemPos + 1
                            Next lBlockCounter
                        Else
                            ' // Just a single ED, write it out
                            mem(lMemPos) = &HED&
                            lMemPos = lMemPos + 1
                        End If
                    Else
                        mem(lMemPos) = Asc(Mid$(sData, lCounter, 1))
                        lMemPos = lMemPos + 1
                    End If
                    lCounter = lCounter + 1
                Loop Until lCounter > Len(sData)
            End If
        End If
    Loop Until EOF(hFile)

End Sub

Sub resetKeyboard()
    keyB_SPC = &HFF&
    keyH_ENT = &HFF&
    keyY_P = &HFF&
    key6_0 = &HFF&
    key1_5 = &HFF&
    keyQ_T = &HFF&
    keyA_G = &HFF&
    keyCAPS_V = &HFF&
End Sub


Public Sub SetDisplaySize(lWidth As Long, lHeight As Long)
    #If USE_WINAPI Then
        Dim i As Integer, Y As Long, X As Long
        
        Y = (GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU) + GetSystemMetrics(SM_CYFRAME) * 2) * Screen.TwipsPerPixelY
        X = (GetSystemMetrics(SM_CXFRAME) * 2) * Screen.TwipsPerPixelX
        glDisplayWidth = lWidth
        glDisplayHeight = lHeight
        glDisplayVSource = lHeight - 1
        glDisplayVSize = -lHeight
        
        SaveSetting "Grok", "vbSpec", "DisplayWidth", CStr(glDisplayWidth)
        SaveSetting "Grok", "vbSpec", "DisplayHeight", CStr(glDisplayHeight)
        
        Select Case lWidth
        Case 256 ' // Standard
            frmMainWnd.Width = 4200 + X
        Case 512
            frmMainWnd.Width = 8040 + X
        Case 768
            frmMainWnd.Width = 11880 + X
        End Select
            
        Select Case lHeight
        Case 192
            frmMainWnd.Height = 3330 + Y
        Case 384
            frmMainWnd.Height = 6210 + Y
        Case 576
            frmMainWnd.Height = 9090 + Y
        End Select
            
        gpicDisplay.Width = lWidth
        gpicDisplay.Height = lHeight
        
        initscreen
    #End If
End Sub


